import json
import time

from sqlglot import exp, parse_one, MappingSchema
from sqlglot.errors import SqlglotError
from sqlglot.lineage import lineage

from dbt_column_lineage.constants import SQLGLOT_DIALECT
from dbt_column_lineage.utils import get_dbt_project_dir


class DbtSqlglot:
    _instance = None

    def __new__(cls, logger, request_depth=-1):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, logger, request_depth=-1):
        # 毎度初期化する変数
        self.nodes = []
        self.edges = []
        self.request_depth = request_depth

        if self._initialized:
            return
        self._initialized = True

        dbt_project_dir = get_dbt_project_dir()
        with open(f'{dbt_project_dir}/target/manifest.json') as f:
            manifest = json.load(f)
        with open(f'{dbt_project_dir}/target/catalog.json') as f:
            catalog = json.load(f)

        self.dbt_metadata = manifest['metadata']
        self.dbt_manifest_nodes = manifest['nodes']
        self.dbt_manifest_sources = manifest['sources']
        self.dbt_catalog_nodes = catalog['nodes']
        self.dbt_manifest_child_map = manifest['child_map']
        self.dbt_manifest_parent_map = manifest['parent_map']

        self.logger = logger
        self.dialect = SQLGLOT_DIALECT

    def project_name(self):
        self.logger.info(self.dbt_metadata)
        return self.dbt_metadata['project_name']

    def schema_by_source(self, source: str):
        dbt_node = self.__get_dbt_node(source)
        return dbt_node.get('schema')

    def columns_by_source(self, source: str):
        dbt_node = self.__get_dbt_node(source)
        dbt_catalog = self.__get_dbt_catalog(source)
        dbt_columns = list(dbt_catalog.get('columns', dbt_node.get('columns', {})).keys())
        lowercase_dbt_columns = [s.lower() for s in dbt_columns]
        return lowercase_dbt_columns

    def list_schemas(self):
        schemas = []
        ret = []
        for k in self.dbt_manifest_nodes:
            dbt_node = self.dbt_manifest_nodes[k]
            dbt_schema = dbt_node['schema']
            if (dbt_schema not in schemas and
                bool(dbt_node['columns']) and
                dbt_node['resource_type'] == 'model' and
                dbt_node['package_name'] == self.dbt_metadata['project_name']):
                schemas.append(dbt_schema)
        schemas.sort()
        for schema in schemas:
            ret.append({'value': schema, 'label': schema})
        return ret

    def list_sources(self, req_schema):
        tmp = []
        for k in self.dbt_manifest_nodes:
            dbt_node = self.dbt_manifest_nodes[k]
            dbt_resource_type = dbt_node['resource_type']
            dbt_schema = dbt_node['schema']
            dbt_alias = dbt_node['alias']
            # (例) fqn = ['data_cuisine_dbt', 'raw_coredb_hourly', 'xxx', 'yyy', 'model_name']
            # ↑の 'xxx/yyy' の部分がほしい
            dbt_fqn = '/'.join(dbt_node['fqn'][2:-1])
            if dbt_resource_type == 'model' and dbt_schema == req_schema:
                tmp.append({'fqn': dbt_fqn, 'alias': dbt_alias})
        label_fqn_alias = {}
        for t in tmp:
            fqn = t['fqn']
            if not label_fqn_alias.get(fqn):
                label_fqn_alias[fqn] = []
            label_fqn_alias[fqn].append({'value': t['alias'], 'label': t['alias']})
        ret = []
        for label in label_fqn_alias:
            # ソート
            label_fqn_alias[label].sort(key=lambda v: v['label'])
            ret.append({'label': label, 'options': label_fqn_alias[label]})
        ret = sorted(ret, key=lambda x:x['label'])
        return ret

    def list_columns(self, req_source):
        for k in self.dbt_manifest_nodes:
            dbt_node = self.dbt_manifest_nodes[k]
            dbt_alias = dbt_node['alias']
            dbt_columns = dbt_node['columns']
            if dbt_alias == req_source:
                ret = []
                for key, value in dbt_columns.items():
                    ret.append({'value': key, 'label': value['name'], 'description': value['description']})
                return ret


    def table_lineage(self, source: str, revs=False):
        dbt_node = self.__get_dbt_node(source)
        unique_id = dbt_node.get('unique_id')
        depth = 0
        self.__table_dependencies_recursive(unique_id, revs, depth)


    def column_lineage(self, source: str, column: str, revs: bool):
        if not revs:
            depth = 0
            self.__column_lineage_recursive('', source, '', [column.upper()], depth)
        else:
            self.__reverse_column_lineage(source, column)


    def ret_edges_nodes(self) -> dict:
        return {'edges': self.edges, 'nodes': self.nodes}


    def cte_dependency(self, source: str, columns: []):
        dbt_node = self.__get_dbt_node(source)
        dbt_catalog = self.__get_dbt_catalog(source)
        table_name = dbt_node.get('name')
        materialized = dbt_node.get('config', {}).get('materialized')
        compiled_code = dbt_node.get('compiled_code')
        description = dbt_node.get('description')
        dbt_columns = dbt_catalog.get('columns')
        dbt_depends_on_nodes = dbt_node.get('depends_on', {}).get('nodes', [])

        if compiled_code is None:
            self.logger.info('compiled_code is None')
            return None
        entire_meta = self.__cte_dependency_impl(dbt_depends_on_nodes, compiled_code, source, columns)
        return {
            'edges': self.edges,
            'nodes': self.nodes,
            'tableName': table_name,
            'materialized': materialized,
            'query': compiled_code,
            'description': description,
            'columns': dbt_columns,
            'entireMeta': entire_meta,
        }

    def __get_dbt_columns(self, dbt_node):
        # dbt_manifest_node の columns がない場合は sources の値を元に dbt_manifest_sources から columns を取得する
        dn_columns = list(dbt_node.get('columns', {}).keys())
        if len(dn_columns) == 0:
            sources = dbt_node.get('sources', [[]])[0]
            source_key = '.'.join(sources)
            # FIXME data_cuisine_dbt は metadata から取る
            ds_columns = list(
                self.dbt_manifest_sources.get(f'source.data_cuisine_dbt.{source_key}', {}).get('columns', {}).keys())
            dn_columns = ds_columns
        return dn_columns


    def __get_depends_on_table_info(self, depends_on_nodes: []) -> []:
        res = []
        for depends_on_node in depends_on_nodes:
            # 依存テーブルから manifest.json を検索
            element = self.dbt_manifest_nodes.get(depends_on_node, self.dbt_manifest_sources.get(depends_on_node))
            if not element:
                continue
            # catalog.json にカラム情報があればそれを使い、なければ manifest.json のカラム情報を使う
            element_columns = element.get('columns', {})
            columns = self.__get_dbt_catalog(element['name']).get('columns', element_columns)
            res.append({
                'columns': columns,
                'table': exp.Table(
                    catalog=exp.Identifier(this=element['database'].upper()),
                    db=exp.Identifier(this=element['schema'].upper()),
                    this=exp.Identifier(this=element['name'].upper()),
                )
            })
        return res

    def __add_node(self, next_source, dbt_schema, filtered_columns, dbt_materialized, depth):
        node_columns = list(filtered_columns)
        node_id = self.__str_to_base_10_int_str(next_source)
        found_node = self.__find(self.nodes, 'id', node_id)

        if found_node:
            # すでにノードがあれば columns を増やす
            exist_columns = found_node['data']['columns']
            for node_column in node_columns:
                # 存在しないカラムだけ追加
                if node_column not in exist_columns:
                    exist_columns.append(node_column)
            # self.logger.info(f'**********already exists node source: {next_source}, columns: {exist_columns}')
            return

        if len(node_columns) != 0:
            max_len = max(max(node_columns, key=len), next_source, key=len)
        else:
            max_len = 0

        # 初回時
        self.nodes.append({
            'id': node_id,
            'data': {
                'name': next_source,
                'materialized' : dbt_materialized,
                'schema': dbt_schema, 'columns': node_columns,
                'first': False, 'last': False,
            },
            'position': {'x': 0,'y': 0},
            # 'max_len': max_len, 'depth': depth,
            'type': 'eventNode'
        })

    def __add_edge(self, base_column: str, columns: [], base_source: str, source: str):
        if not base_source or len(columns) == 0:
            self.logger.debug(f'{base_source} is None or {columns} len = 0')
            return
        s = self.__str_to_base_10_int_str(base_source)
        t = self.__str_to_base_10_int_str(source.lower())
        base_edge_id = f'{s}-{t}'
        # if not (self.find(self.edges, 'source', str(s)) and self.find(self.edges, 'target', str(t))):
        for column in columns:
            edge_id = f'{base_edge_id}-{base_column}-{column}'
            if self.__find(self.edges, 'id', edge_id):
                self.logger.debug(f'already exist edge: {edge_id}')
                continue
            else:
                self.logger.debug(f'edge:source={base_source}, target={source.lower()}, base_column={base_column}, column={column}')
                self.edges.append({
                    'id': edge_id,
                    'source': str(s),
                    'target': str(t),
                    'sourceHandle': f'{base_column}__source',
                    'targetHandle': f'{column}__target',
                })

    def __get_parent_node(self, next_source, depth):
        for n in reversed(self.nodes):
            prev_name = n['data']['name']
            if prev_name != next_source and n['depth'] == depth - 1:
                self.logger.info('parent', prev_name, next_source, depth)
                return n
        return None

    def __get_sibling_node(self, next_source, depth):
        for n in reversed(self.nodes):
            prev_name = n['data']['name']
            if prev_name != next_source and n['depth'] == depth:
                self.logger.info('sibling', prev_name, next_source, depth)
                return n
        return None

    def __get_columns(self, dbt_columns: [], next_columns: []) -> []:
        if len(dbt_columns) == 0:
            self.logger.error('dbt_columns is empty')
            return next_columns

        ret = []
        # 小文字に変換
        dbt_columns = [x.lower() for x in dbt_columns]
        # next_column が dbt のカラムに含まれているものだけ返す
        for next_column in next_columns:
            if next_column.lower() in dbt_columns:
                ret.append(next_column)

        return ret

    def __get_dbt_node(self, dbt_target: str) -> {}:
        for resource_type in ['model', 'seed', 'snapshot']:
            # FIXME data_cuisine_dbt は metadata から取る
            key = f'{resource_type}.data_cuisine_dbt.{dbt_target}'
            element = self.dbt_manifest_nodes.get(key, {})
            # self.logger.info(key, len(element))
            if len(element) > 0:
                break
        return element

    def __get_dbt_catalog(self, dbt_target: str) -> {}:
        for resource_type in ['model', 'seed', 'snapshot']:
            key = f'{resource_type}.data_cuisine_dbt.{dbt_target}'
            element = self.dbt_catalog_nodes.get(key, {})
            if len(element) > 0:
                break
        return element

    def __get_sqlglot_lineage(self, source: str, compiled_code: str, columns: [], depends_on_table_info: [], need_meta=False) -> dict:
        # MappingSchema にテーブル情報を追加
        sqlglot_db_schema = MappingSchema(dialect=self.dialect, normalize=False)
        for s in depends_on_table_info:
            source_table = s['table']
            source_columns = s['columns']

            table_schema = {}
            for key, item in source_columns.items():
                table_schema[key] = item.get('type', 'STRING')

            sqlglot_db_schema.add_table(
                source_table,
                column_mapping=table_schema,
            )
        try:
            parsed_sql = parse_one(compiled_code, dialect=self.dialect)
        except SqlglotError:
            self.logger.error(f'parse sql. source={source}')
            return {}

        ret = {}
        for column in columns:
            column = column.upper()
            labels = set()
            replace_columns = set()
            try:
                start_time = time.time()
                lin = lineage(column, parsed_sql, dialect=self.dialect, schema=sqlglot_db_schema)
                end_time = time.time()
                if end_time - start_time > 3:
                    self.logger.info(f'lineage time: source={source}, column={column}, time={end_time - start_time}秒')
            except SqlglotError:
                self.logger.error(f'lineage error. source={source}, column={column}')
                continue

            meta = []
            for node in lin.walk():
                if isinstance(node.expression, exp.Table):
                    label = f'{node.expression.this}'
                    # 配下のデータがなければ最後とみなす
                    self.logger.debug(f'label: {label}')
                    if len(node.downstream) == 0:
                        labels.add(label)
                if node.name != '*' and not isinstance(node.expression, exp.Table):
                    cte = node.expression.sql(dialect=self.dialect)
                    self.logger.debug(cte)
                    try:
                        parsed_sql = parse_one(cte, dialect=self.dialect)
                        cl = parsed_sql.find_all(exp.Column)
                    except SqlglotError:
                        self.logger.error(f'cte parse error. source={source}, column={column}, cte={cte}')
                        cl = []
                    for c in cl:
                        self.logger.debug(f'alias_or_name={c.alias_or_name}')
                        replace_columns.add(c.alias_or_name)
                # CTEリネージ用の追加情報
                if need_meta and not isinstance(node.expression, exp.Table):
                    if node.reference_node_name == '':
                        # 最初のselect * from finalは無視
                        continue

                    node_downstream_alias_names = []
                    node_downstream_table_sources = []
                    for d in node.downstream:
                        if isinstance(d.expression, exp.Table):
                            node_downstream_table_sources.append({'schema': d.expression.db.lower(),'table': d.expression.name.lower()})
                            parsed_column = parse_one(d.name, dialect=self.dialect)
                            node_downstream_alias_names.append(parsed_column.alias_or_name.lower())
                        else:
                            node_downstream_alias_names.append(d.expression.alias_or_name.lower())

                    # カラム名が一致したら次のカラムをリセット
                    column_name = node.expression.alias_or_name.lower()
                    if len(node_downstream_alias_names) > 0 and column_name == node_downstream_alias_names[0]:
                        node_downstream_alias_names = []
                    mt = {
                        'column': node.expression.alias_or_name.lower(),
                        'nextColumns': node_downstream_alias_names,
                        'nextSources': node_downstream_table_sources,
                        'reference': node.reference_node_name.lower()
                    }
                    meta.append(mt)

            ret[column] = {'labels': list(labels), 'columns': list(replace_columns)}
            if need_meta:
                ret[column]['meta'] = meta
        self.logger.info(f'{source}, {ret}')
        return ret

    def __str_to_base_10_int_str(self, s:str) -> str:
        hash_value = 0
        for char in s:
            hash_value = (hash_value << 5) - hash_value + ord(char)
            hash_value = hash_value & 0xFFFFFFFF  # Convert to 32-bit integer
        return str(abs(hash_value))

    def __find(self, arr: [], key: str, value: str):
        for x in arr:
            if x[key] == value:
                return x

    def __find_all(self, arr: [], key: str, value: str):
        ret = []
        for x in arr:
            if x[key] == value:
                ret.append(x)
        return ret

    def __find_with_subkey(self, arr: [], key: str, subkey: str, value: str):
        for x in arr:
            if x[key][subkey] == value:
                return x

    def __get_table_dependencies(self, reverse, unique_id):
        if reverse:
            refs = self.dbt_manifest_child_map.get(unique_id, [])
        else:
            refs = self.dbt_manifest_parent_map.get(unique_id, [])
        return refs

    def __table_dependencies_recursive(self, unique_id, reverse, depth):
        self.logger.info(f'request_depth={self.request_depth}, depth={depth}')
        ref_dbt_node = self.dbt_manifest_nodes.get(unique_id)
        if ref_dbt_node is None:
            self.logger.error(f'unique_id={unique_id} is not found')
            return
        ref_unique_id = ref_dbt_node.get('unique_id')
        ref_name = ref_dbt_node.get('name')
        ref_schema = ref_dbt_node.get('schema')
        ref_materialized = ref_dbt_node.get('config', {}).get('materialized')
        deps_refs= self.__get_table_dependencies(reverse, ref_unique_id)
        node_id = self.__str_to_base_10_int_str(ref_name)

        if not self.__find(self.nodes, 'id', node_id):
            self.nodes.append({
                'id': node_id,
                'data': {
                    'name': ref_name,
                    'columns': [],
                    'schema': ref_schema,
                    'materialized': ref_materialized,
                    'first': False, 'last': False
                },
                'position': {'x': 0,'y': 0},
                # 'max_len': max_len,'depth': depth,
                'type': 'eventNode'
            })

        depth = depth + 1
        if self.request_depth != -1 and depth > self.request_depth:
            self.logger.info(f'depth={depth} reached')
            return

        for deps_unique_id in deps_refs:
            self.logger.info(deps_unique_id)
            deps_ref_dbt_node = self.dbt_manifest_nodes.get(deps_unique_id)
            if deps_ref_dbt_node is None:
                self.logger.error(f'unique_id={deps_unique_id} is not found')
                return
            target_name = deps_ref_dbt_node.get('name')
            target_node_id = self.__str_to_base_10_int_str(target_name)
            if reverse:
                edge_id = f'{target_node_id}-{node_id}'
                if not self.__find(self.edges, 'id', edge_id):
                    self.edges.append({
                        'id': edge_id,
                        'source': target_node_id,
                        'target': node_id,
                        'sourceHandle': f'{target_node_id}__source',
                        'targetHandle': f'{node_id}__target'
                    })
            else:
                edge_id = f'{node_id}-{target_node_id}'
                if not self.__find(self.edges, 'id', edge_id):
                    self.edges.append({
                        'id': edge_id,
                        'source': node_id,
                        'target': target_node_id,
                        'sourceHandle': f'{node_id}__source',
                        'targetHandle': f'{target_node_id}__target'
                    })
            self.__table_dependencies_recursive(deps_unique_id, reverse, depth)
        self.logger.info(f'end')
        return

    def __column_lineage_recursive(self, base_source: str, next_source: str, base_column: str, next_columns: [], depth: int) -> None:
        dbt_catalog = self.__get_dbt_catalog(next_source)
        dbt_node = self.__get_dbt_node(next_source)
        dbt_compiled_code = dbt_node.get('compiled_code')
        dbt_schema = dbt_node.get('schema')
        dbt_depends_on_nodes = dbt_node.get('depends_on', {}).get('nodes', [])
        dbt_columns = list(dbt_catalog.get('columns', dbt_node.get('columns', {})).keys())
        dbt_materialized = dbt_node.get('config', {}).get('materialized')

        # dbt の定義に該当するカラムだけに絞り込む
        filtered_columns = self.__get_columns(dbt_columns, next_columns)

        # ノード作成
        self.__add_node(next_source, dbt_schema, filtered_columns, dbt_materialized, depth)
        # エッジ作成
        self.__add_edge(base_column, filtered_columns, base_source, next_source)

        if dbt_compiled_code is None:
            self.logger.info('dbt_compiled_code is None')
            after_next_sources_columns_dict = {}
        else:
            # リネージの手がかりとして依存テーブルのカラムやテーブル情報を作成
            additional_dbt_sources = self.__get_depends_on_table_info(dbt_depends_on_nodes)

            after_next_sources_columns_dict = self.__get_sqlglot_lineage(next_source, dbt_compiled_code, filtered_columns, additional_dbt_sources)

        depth = depth + 1
        after_base_source = next_source
        next_found = False
        for after_base_column, after_next_sources_columns in after_next_sources_columns_dict.items():
            if self.request_depth != -1 and depth > self.request_depth:
                continue
            after_next_sources = after_next_sources_columns['labels']
            after_next_columns = after_next_sources_columns['columns']
            for after_next_source in after_next_sources:
                next_found = True
                after_next_source = after_next_source.lower()

                self.logger.debug(f'base_source={after_base_source}, next_source={after_next_source}, base_column={after_base_column}, next_columns={after_next_columns}, depth={depth}')
                self.__column_lineage_recursive(after_base_source, after_next_source, after_base_column, after_next_columns, depth)
        if not next_found:
            # 再起の最後だったらedgeの起点がない
            prev_node = self.__find_with_subkey(self.nodes, 'data', 'name', after_base_source)
            if prev_node and self.request_depth == -1:
                self.logger.debug(f'last node: {after_base_source}')
                prev_node['data']['last'] = True

    def __reverse_column_lineage(self, source: str, column: str):
        dbt_node = self.__get_dbt_node(source)
        unique_id = dbt_node.get('unique_id')
        refs = self.dbt_manifest_child_map.get(unique_id, [])

        data = {}
        for ref in refs:
            ref_dbt_node = self.dbt_manifest_nodes.get(ref)
            ref_node_name = ref_dbt_node.get('name')
            ref_schema = ref_dbt_node.get('schema')
            ref_compiled_code = ref_dbt_node.get('compiled_code')
            ref_dbt_depends_on_nodes = ref_dbt_node.get('depends_on', {}).get('nodes', [])
            ref_materialized = ref_dbt_node.get('config', {}).get('materialized')

            depends_on_table_info = self.__get_depends_on_table_info(ref_dbt_depends_on_nodes)
            found = False
            for info in depends_on_table_info:
                tbl : exp.Table = info['table']
                if tbl.name.lower() == source:
                    print(tbl.name)
                    found = True
            if not found:
                continue

            # catalog.json にカラム情報があればそれを使い、なければ manifest.json のカラム情報を使う
            ref_columns = ref_dbt_node.get('columns', self.__get_dbt_catalog(ref).get('columns', {}))
            for ref_column in ref_columns:
                self.logger.info(f'table={ref}, column={ref_column}')

                ref_column_name = ref_column.upper()
                items = self.__get_sqlglot_lineage(source, ref_compiled_code, [ref_column_name], depends_on_table_info)
                item_labels_columns = items.get(ref_column_name, {})
                if column in item_labels_columns.get('columns', []):
                    item_labels = item_labels_columns['labels']
                    item_columns = item_labels_columns['columns']
                    if source.upper() in item_labels and column in item_columns:
                        self.logger.info(f'source={source}, column={column} found')
                        r = data.get(ref_node_name, {'columns': [], 'schema': ref_schema, 'materialized': ref_materialized})
                        r['columns'].append(ref_column_name)
                        data[ref_node_name] = r

        ret_nodes = []
        ret_edges = []
        target_id = self.__str_to_base_10_int_str(source)
        for node_name, item in data.items():
            target_schema = item['schema']
            target_columns = item['columns']
            target_materialized = item['materialized']
            node_id = self.__str_to_base_10_int_str(node_name)
            base_edge_id = f'{node_id}-{target_id}'
            ret_nodes.append({
                'id': node_id,
                'data': {
                    'name': node_name,
                    'schema': target_schema,
                    'columns': target_columns,
                    'materialized': target_materialized,
                    'first': False,'last': False
                },
                'position': {'x': 0,'y': 0},
                'type': 'eventNode'
            })
            for target_column in target_columns:
                edge_id = f'{base_edge_id}-{target_column}-{column}'
                ret_edges.append({
                    'id': edge_id,
                    'source': node_id,
                    'target': target_id,
                    'sourceHandle': f'{target_column}__source',
                    'targetHandle': f'{column}__target'
                })
        self.edges = ret_edges
        self.nodes = ret_nodes

    def __cte_dependency_impl(self, dbt_depends_on_nodes: [], compiled_code: str, source: str, columns: []) -> list:
        dependencies = {}
        lineage_tables = []
        lineage_meta = []

        if len(columns) > 0:
            depends_on_table_info = self.__get_depends_on_table_info(dbt_depends_on_nodes)
            items = self.__get_sqlglot_lineage(source, compiled_code, columns, depends_on_table_info, True)

            # 現状columns は1つのみ
            item = items.get(columns[0].upper(), {'labels': [], 'meta': {}})
            for label in item['labels']:
                lineage_tables.append(label.lower())
            lineage_meta = item['meta']

        try:
            parsed_sql = parse_one(compiled_code, dialect=self.dialect)
            ctes = parsed_sql.find_all(exp.CTE)
        except SqlglotError:
            self.logger.error(f'parse cte error. source={source}')
            ctes = []

        for cte in ctes:
            dependencies[cte.alias_or_name] = []
            # reference が一致するすべての meta 情報を取得
            meta = self.__find_all(lineage_meta, 'reference', cte.alias_or_name)

            # CTE の各要素を抽出
            elements = {
                'groups': [ele.sql() for ele in cte.find_all(exp.Group)],
                'havings': [ele.sql() for ele in cte.find_all(exp.Having)],
                'wheres': [ele.sql() for ele in cte.find_all(exp.Where)],
                'unions': [ele.sql() for ele in cte.find_all(exp.Union)],
                'joins': [ele.sql() for ele in cte.find_all(exp.Join)]
            }

            # ノードデータの作成
            node_data = {
                'id': cte.alias_or_name,
                'type': 'cte',
                'data': {
                    'label': cte.alias_or_name,
                    'nodeType': 'CTE',
                    'meta': meta,
                    **elements
                },
                'position': {'x': 0, 'y': 0},
            }

            self.nodes.append(node_data)

            query = cte.this.sql()
            try:
                parsed_cte_query = parse_one(query, dialect=self.dialect)
                tables = parsed_cte_query.find_all(exp.Table)
            except SqlglotError:
                self.logger.error(f'parse cte sql error. source={source}, query={query}')
                tables = []

            for table in tables:
                dependencies[cte.alias_or_name].append({'name': table.name})

        for node in dependencies:
            for dependency in dependencies[node]:
                dep = dependency['name']
                edge_id = f'{node}-{dep}'
                if not self.__find(self.edges, 'id', edge_id) and node and dep and node != dep:
                    self.edges.append({'id': edge_id, 'source': dep,'target': node, 'markerStart': {'type': 'arrowclosed', 'width': 16, 'height': 16}})

        return lineage_meta




import looker_sdk
from looker_sdk.sdk.api40.models import Folder


class Looker:
    def __init__(self, logger):
        self.logger = logger
        self.sdk = looker_sdk.init40()
        self.ignore_folder_names = ['tmp', 'old', 'バックアップ', '作業', 'テスト']
        self.ignore_dashboard_element_titles = ['表示期間', '無題', 'データ取込日時']

    def get_folder_dashboards(self) -> []:
        self.folders = {}
        res = self.sdk.all_folders(fields='id, parent_id, name, is_personal, is_personal_descendant, child_count')
        for r in res:
            # 個人用フォルダ関連は無視
            if r.is_personal or r.is_personal_descendant:
                continue
            self.folders[r.id] = r

        ret = []
        for v in self.folders.values():
            # 子フォルダがある場合は無視
            if v.child_count:
                continue
            # 再帰的にフォルダ名を取得
            name = self._get_folder(v, v.name)
            # 共有フォルダ以外無視
            if not name.startswith('Shared'):
                continue
            ignore = False
            # ignore_folder_names に含まれる文字列を含むフォルダは無視（「tmp」など）
            for ifn in self.ignore_folder_names:
                if ifn in name:
                    ignore = True
                    break
            if ignore:
                continue

            # フォルダIDからダッシュボード一覧を取得
            res = self.sdk.folder_dashboards(folder_id=v.id, fields='id,title')
            dashboards = []
            for d in res:
                dashboards.append({'id': d.id, 'name': d.title})
            ret.append({'id': v.id, 'name': name, 'dashboards': dashboards})
        return ret

    def get_dashboard_elements(self, dashboard_id: str) -> []:
        dashboard = self.sdk.dashboard(dashboard_id=dashboard_id, fields='dashboard_elements,dashboard_layouts')
        dashboard_layout_components = []
        for item in dashboard.dashboard_layouts:
            for layout_component in item.dashboard_layout_components:
                dashboard_layout_components.append(layout_component)

        # dashboard_layout の row, column でソート
        sorted_dashboard_layout_components = sorted(dashboard_layout_components, key=lambda x: (x.row, x.column))
        dashboard_elements = {}
        for item in dashboard.dashboard_elements:
            dashboard_elements[item.id] = item

        ret = []
        # ソートされた layout_components から dashboard_element を取得
        for layout_component in sorted_dashboard_layout_components:
            item = dashboard_elements[layout_component.dashboard_element_id]
            # ignore_dashboard_element_titles に含まれる文字列を含む要素は無視 (「無題」など)
            ignore = False
            for idt in self.ignore_dashboard_element_titles:
                if idt in item.title:
                    ignore = True
                    break
            if ignore:
                continue

            # result_maker が存在しない場合は無視
            if item.result_maker is None:
                continue

            ret.append({
                'element_id': item.id,
                'title': item.title,
                'share_url': item.result_maker.query.share_url,
                'slug': item.result_maker.query.slug,
            })

        return ret

    def get_explore_fields(self, slug: str) -> []:
        query = self.sdk.query_for_slug(slug)
        res = self.sdk.lookml_model_explore(lookml_model_name=query.model, explore_name=query.view,
                                            fields='fields, sql_table_name')

        # table_name が AS で別名がついている場合は元のテーブル名を取得
        if 'AS' in res.sql_table_name:
            table_name = res.sql_table_name.split('AS')[0].strip()
        else:
            table_name = res.sql_table_name

        # schema.table の場合は table のみ取得
        if '.' in table_name:
            table_name = table_name.split('.')[1]

        query_sql = self.sdk.run_query(query.id, result_format='sql')
        ret = {
            'id': res.id,
            'name': res.name,
            'title': res.title,
            'sql': query_sql,
            'table_name': table_name,
            'dimensions': [],
            'measures': [],
            'filters': []
        }

        for dimension in res.fields.dimensions:
            if query.fields and dimension.name in query.fields:
                dim_def = {
                    'field_type': 'dimension',
                    'name': dimension.name,
                    'view_label': dimension.view_label,
                    'label_short': dimension.label_short,
                    'type': dimension.type,
                    'sql': dimension.sql,
                }
                ret['dimensions'].append(dim_def)
            elif query.filters and dimension.name in query.filters:
                filter_def = {
                    'field_type': 'filter',
                    'name': dimension.name,
                    'view_label': dimension.view_label,
                    'label_short': dimension.label_short,
                    'type': dimension.type,
                    'sql': dimension.sql
                }
                ret['filters'].append(filter_def)

        for measure in res.fields.measures:
            if query.fields and measure.name in query.fields:
                mes_def = {
                    'field_type': 'measure',
                    'name': measure.name,
                    'view_label': measure.view_label,
                    'label_short': measure.label_short,
                    'type': measure.type,
                    'sql': measure.sql
                }
                ret['measures'].append(mes_def)

        return ret


    def _get_folder(self, folder: Folder, folder_name: str) -> str:
        if folder and folder.parent_id:
            parent_folder = self.folders[folder.parent_id]
            folder_name = f'{parent_folder.name}/{folder_name}'
            return self._get_folder(parent_folder, folder_name)
        return folder_name




import base64
import hashlib
import json
import os
import urllib
from contextlib import asynccontextmanager
from pathlib import Path
from urllib import parse
from urllib.parse import urlencode

import requests
from fastapi import FastAPI, HTTPException, Depends, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.params import Query
from fastapi.responses import JSONResponse, RedirectResponse, FileResponse
from fastapi.security import OAuth2PasswordBearer
from fastapi.staticfiles import StaticFiles
from requests.auth import HTTPBasicAuth
from starlette.middleware.sessions import SessionMiddleware

from dbt_column_lineage.constants import USE_OAUTH, GOOGLE_CLIENT_ID, GOOGLE_CLIENT_SECRET, BASE_ROUTE
from dbt_column_lineage.lineage import DbtSqlglot
from dbt_column_lineage.looker import Looker
from dbt_column_lineage.utils import get_logger, get_redirect_url, get_diff_to_params, startup_dialect_check

import typer

cli = typer.Typer()

@asynccontextmanager
async def lifespan(app: FastAPI):
    startup_dialect_check()
    yield

app = FastAPI(lifespan=lifespan)

# CORS設定
app.add_middleware(CORSMiddleware, allow_origins=['*'], allow_credentials=True, allow_methods=['*'], allow_headers=['*'])

# セッション管理のミドルウェア追加
app.add_middleware(SessionMiddleware, secret_key=os.urandom(32).hex())

logger = get_logger(app, __name__)

oauth2_scheme = OAuth2PasswordBearer(tokenUrl='token')


# セッションからアクセストークンを取得する依存関数
async def get_current_user(request: Request):
    if USE_OAUTH:
        access_token = request.session.get('access_token')
        if not access_token:
            raise HTTPException(status_code=401, detail='Not authenticated')
        return access_token
    return None


@app.get('/healthcheck')
async def readiness_probe():
    return 'ok!'


@app.exception_handler(404)
async def not_found(request: Request, exc: HTTPException):
    return FileResponse(f'frontend_out/404.html')


@app.get('/')
@app.get('/cl')
@app.get('/cte')
@app.get('/login')
async def static_page(request: Request):
    request_path = request.url.path
    if USE_OAUTH:
        if request_path == '/login':
            request.session['csrf'] = os.urandom(32).hex()
        elif 'access_token' not in request.session:
            return RedirectResponse(url='/login')
        logger.debug(f'access_token: {request.session.get("access_token")}')

    if request_path == '/':
        static_path = 'index'
    else:
        static_path = request_path[1:]

    APP_ROOT = os.path.dirname(os.path.abspath(__file__))
    return FileResponse(f'{APP_ROOT}/frontend_out/{static_path}.html')


@app.get('/oauth')
async def oauth(request: Request):
    if 'csrf' not in request.session:
        return RedirectResponse(url='/login')

    authorization_endpoint = 'https://accounts.google.com/o/oauth2/v2/auth'
    scope = 'openid https://www.googleapis.com/auth/userinfo.email'

    code_verifier = os.urandom(43).hex()
    code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode('utf-8')).digest()).decode('utf-8').replace('=','')
    request.session['code_verifier'] = code_verifier

    params = {
        'response_type': 'code',
        'client_id': GOOGLE_CLIENT_ID,
        'state': request.session.get('csrf'),
        'scope': scope,
        'redirect_uri': get_redirect_url(request),
        'code_challenge': code_challenge,
        'code_challenge_method': 'S256'
    }

    return RedirectResponse(url=f'{authorization_endpoint}?{urlencode(params)}')


@app.get('/callback')
async def callback(request: Request, state: str, code: str):
    if ('csrf' not in request.session) or (request.session['csrf'] != state):
        return RedirectResponse(url='/')
    if 'code_verifier' not in request.session:
        return RedirectResponse(url='/')

    params = {
        'client_id': GOOGLE_CLIENT_ID,
        'client_secret': GOOGLE_CLIENT_SECRET,
        'redirect_uri': get_redirect_url(request),
        'grant_type': 'authorization_code',
        'code': code,
        'code_verifier': request.session['code_verifier']
    }

    headers = {'content-type': 'application/x-www-form-urlencoded'}
    response = requests.post(
        'https://oauth2.googleapis.com/token',
        params=params,
        headers=headers,
        auth=HTTPBasicAuth(GOOGLE_CLIENT_ID, GOOGLE_CLIENT_SECRET)
    )

    data = response.json()
    logger.debug(data)
    access_token = data['access_token']
    request.session['access_token'] = access_token

    return RedirectResponse(url='/')


@app.get(f'{BASE_ROUTE}/schemas')
async def list_schemas(current_user: str = Depends(get_current_user)):
    dbt_sqlglot = DbtSqlglot(logger)
    schemas = dbt_sqlglot.list_schemas()
    return JSONResponse(content=schemas)


@app.get(f'{BASE_ROUTE}/sources')
async def list_sources(schema: str, current_user: str = Depends(get_current_user)):
    dbt_sqlglot = DbtSqlglot(logger)
    models = dbt_sqlglot.list_sources(schema)
    return JSONResponse(content=models)


@app.get(f'{BASE_ROUTE}/columns')
async def list_columns(source: str, current_user: str = Depends(get_current_user)):
    dbt_sqlglot = DbtSqlglot(logger)
    columns = dbt_sqlglot.list_columns(source)
    return JSONResponse(content=columns)


@app.get(f'{BASE_ROUTE}/lineage')
async def find_lineage(
    sources: str = Query(..., description='source table names'),
    columns: str = Query(..., description='columns name'),
    show_column: bool = Query(False, description='show column lineage'),
    reverse: bool = Query(False, description='reverse lineage'),
    depth: int = Query(-1, description='depth of lineage', ge=-1),
    current_user: str = Depends(get_current_user)
):
    dbt_sqlglot = DbtSqlglot(logger, request_depth=depth)
    if show_column:
        parsed_columns = json.loads(columns)
        for source in sources.split(','):
            for column in parsed_columns.get(source, ['']):
                dbt_sqlglot.column_lineage(source, column.upper(), reverse)
    else:
        for source in sources.split(','):
            dbt_sqlglot.table_lineage(source, reverse)
    res = dbt_sqlglot.ret_edges_nodes()
    return JSONResponse(content=res)


@app.get(f'{BASE_ROUTE}/cte')
async def find_cte(source: str, column: str = None, current_user: str = Depends(get_current_user)):
    columns = column.split(',') if column and column != 'null' else []
    logger.info(columns)
    dbt_sqlglot = DbtSqlglot(logger)
    res = dbt_sqlglot.cte_dependency(source, columns)
    if res is None:
        raise HTTPException(status_code=404, detail='Not found')
    return JSONResponse(content=res)


@app.get(f'{BASE_ROUTE}/folder_dashboards')
async def list_folder_dashboards(current_user: str = Depends(get_current_user)):
    looker = Looker(logger)
    folder_dashboards = looker.get_folder_dashboards()
    return JSONResponse(content=folder_dashboards)


@app.get(f'{BASE_ROUTE}/dashboard_elements/{{dashboard_id}}')
async def get_dashboard_elements(dashboard_id: str, current_user: str = Depends(get_current_user)):
    looker = Looker(logger)
    dashboard_elements = looker.get_dashboard_elements(dashboard_id)
    return JSONResponse(content=dashboard_elements)


@app.get(f'{BASE_ROUTE}/explore_fields/{{slug}}')
async def get_explore_fields(slug: str, current_user: str = Depends(get_current_user)):
    looker = Looker(logger)
    explore_fields = looker.get_explore_fields(slug)
    return JSONResponse(content=explore_fields)


# 静的ファイルの設定
static_files_path = Path(__file__).parent / 'frontend_out'
app.mount('/', StaticFiles(directory=str(static_files_path), html=True), name='static')

@cli.command()
def run_params():
    dbt_sql_glot = DbtSqlglot(logger)
    sources, selected_columns = get_diff_to_params(dbt_sql_glot)

    if len(sources) != 0:
        first_schema = dbt_sql_glot.schema_by_source(sources[0])
        converted_sources = f'schema={first_schema}&sources={parse.quote(",".join(sources))}'
        show_column = 'false'

        if len(selected_columns) != 0:
            converted_columns = urllib.parse.quote(json.dumps(selected_columns))
            additional_info = f'&activeSource={sources[0]}&selectedColumns={converted_columns}'
            converted_sources += additional_info

        print(f'http://127.0.0.1:5000/cl?{converted_sources}&showColumn={show_column}&depth=0')
    else:
        print('model files diff not found')
    run()

@cli.command()
def run(host: str = '127.0.0.1', port: int = 5000):
    import uvicorn
    uvicorn.run('dbt_column_lineage.main:app', host=host, port=port)

@cli.command()
def version():
    from dbt_column_lineage._version import __version__
    typer.echo(f'dbt-column-lineage version {__version__}')


def main():
    cli()


if __name__ == '__main__':
    main()



import logging
import os
import re
import subprocess
import sys
from datetime import datetime
from pathlib import Path

from fastapi import Request
from pytz import timezone
from sqlglot import parse_one, diff
from sqlglot.diff import Keep, Insert, Update
from sqlglot.expressions import CTE, Table, Literal

from dbt_column_lineage.constants import DEBUG_MODE


def get_logger(app, logger_name):
    logger = logging.getLogger(logger_name)
    handler = logging.StreamHandler(sys.stdout)

    if app.debug or DEBUG_MODE:
        logger.setLevel(logging.DEBUG)
        handler.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
        handler.setLevel(logging.INFO)

    # apprunner では不要説
    formatter = logging.Formatter('%(name)s [%(asctime)s] [%(levelname)s] %(filename)s:%(lineno)d %(message)s')
    formatter.converter = custom_timezone_jst
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def get_redirect_url(request: Request):
    host_url = str(request.base_url)
    if request.url.scheme == 'http' and 'localhost' not in host_url:
        host_url = host_url.replace('http://', 'https://')
    return host_url + 'callback'


def find_dbt_project() -> object:
    # 1. カレントディレクトリをチェック
    current_dir = Path.cwd()
    if (current_dir / 'dbt_project.yml').exists():
        return str(current_dir)

    # 2. ユーザーのホームディレクトリをチェック
    home_dir = Path.home()
    common_locations = [
        'dbt_projects',
        'projects/dbt',
        '.dbt'
    ]
    for location in common_locations:
        path = home_dir / location
        if (path / 'dbt_project.yml').exists():
            return str(path)

    # 3. システム全体の一般的な場所をチェック
    system_locations = [
        '/etc/dbt',
        '/var/lib/dbt',
        '/opt/dbt'
    ]
    for location in system_locations:
        path = Path(location)
        if (path / 'dbt_project.yml').exists():
            return str(path)

    # 4. 見つからない場合は None を返す
    return None


def get_dbt_project_dir():
    # 環境変数をチェック
    dbt_project_dir = os.getenv('DBT_PROJECT_DIR')
    if dbt_project_dir:
        return dbt_project_dir

    # 自動検出を試みる
    detected_dir = find_dbt_project()
    if detected_dir:
        return detected_dir

    # 見つからない場合はエラーを発生させる
    raise ValueError("DBT project directory not found. Please set the DBT_PROJECT_DIR environment variable or run the command from a dbt project directory.")


def custom_timezone_jst(*args):
    return datetime.now(timezone('Asia/Tokyo')).timetuple()


def get_diff_to_params(dbt_sql_glot):
    # プロジェクト名を取得
    project_name = dbt_sql_glot.project_name()

    # 変更されたファイルのリストを取得
    changed_files = _get_git_changed_files()

    # gitワークツリーからの相対パスを取得
    relative_path = _get_relative_path()

    sources = []
    selected_columns = {}

    for change_file in changed_files:
        print(change_file)
        file = os.path.relpath(change_file, relative_path)

        if not os.path.isfile(file):
            print(f'{file} does not exist')
            sys.exit(1)

        source = _extract_model_name(file)
        if source:
            sources.append(source)
            # 既存のカラム情報を取得
            exists_columns = dbt_sql_glot.columns_by_source(source)
            # クエリの差分からカラム情報を取得
            columns = _source_to_columns(file, project_name, dbt_sql_glot.dialect, exists_columns)
            if len(columns) > 0:
                selected_columns[source] = columns

    return sources, selected_columns


def _extract_model_name(input_path):
    # パスを分解
    parts = input_path.split(os.sep)

    # 'models' または 'snapshots' ディレクトリ以下のパスかどうかを確認
    target_dirs = ['models', 'snapshots']
    for target_dir in target_dirs:
        if target_dir in parts:
            dir_index = parts.index(target_dir)
            if len(parts) > dir_index + 1:
                # ファイル名を取得（拡張子を除く）
                filename = parts[-1]
                # ファイル名が 'audit_' で始まる場合は無視
                if filename.startswith('audit_'):
                    continue
                # sqlファイルの場合はファイル名を返す
                if filename.endswith('.sql'):
                    filename = os.path.splitext(filename)[0]
                    return filename

    return None

def _get_git_changed_files():
    command = [
        'git',
        'diff',
        '--no-renames',
        '--name-only',
        '--diff-filter=ACMRT'
    ]

    try:
        # git diffコマンドを実行
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        changed_files = result.stdout.strip().split('\n')
        # 空の行を除去（最後の改行による空行への対応）
        changed_files = [file for file in changed_files if file]

        return changed_files
    except subprocess.CalledProcessError as e:
        print(f'An error occurred while running git command: {e}')
        return []


def _get_relative_path():
    git_root = subprocess.check_output(['git', 'rev-parse', '--show-toplevel'],
                                       universal_newlines=True).strip()
    if git_root is None:
        print('Not in a Git repository')
        sys.exit(1)

    # カレントディレクトリの相対パスを取得
    current_path = os.getcwd()
    return os.path.relpath(current_path, git_root)


def _replace_file_path(original_path, project_name, target_dir='target'):
    # ファイルパスのパターンを定義
    old_pattern = r'^(.*)(models|snapshots)/(.+)$'
    new_pattern = f'\\1{target_dir}/compiled/{project_name}/\\2/\\3'

    # パスを置換
    return re.sub(old_pattern, new_pattern, original_path)


def _source_to_columns(file, project_name, dialect, exists_columns):
    columns = []
    target_path = _replace_file_path(file, project_name, 'target')
    target_base_path = _replace_file_path(file, project_name, 'target-base')

    if not os.path.isfile(target_path):
        print(f'{target_path} does not exist')
        return columns

    if not os.path.isfile(target_base_path):
        print(f'{target_base_path} does not exist')
        return columns

    target_sql = open(target_path, 'r').read()
    target_base_sql = open(target_base_path, 'r').read()

    parsed_sql = parse_one(target_sql, dialect=dialect)
    changes = diff(parse_one(target_base_sql, dialect=dialect), parsed_sql)

    for change in changes:
        if isinstance(change, Keep):
            continue

        # print(change)
        if (isinstance(change, Update)
            and not isinstance(change.target, Table) and not isinstance(change.target, Literal)):
            print('Update:')
            print(change.target)

            if change.target.alias_or_name in exists_columns:
                columns.append(change.target.alias_or_name)
        elif isinstance(change, Insert):
            print('Insert:')
            print(change)

            for cte in parsed_sql.find_all(CTE):
                if cte.sql() == change.expression.parent.parent.sql():
                    if change.expression.alias_or_name in exists_columns:
                        print(change.expression.alias_or_name)
                        columns.append(change.expression.alias_or_name)
                        break
    return columns

def startup_dialect_check():
    if 'SQLGLOT_DIALECT' not in os.environ:
        dialect = 'snowflake'
        print(f'Warning: SQLGLOT_DIALECT not set. Using default: {dialect}')



import warnings
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version('dbt_column_lineage')
except PackageNotFoundError:
    warnings.warn('Failed to read version. Make sure `setuptools_scm` is installed and its setup is called.')



import os

BASE_ROUTE = '/api/v1'
DEBUG_MODE = os.getenv('DEBUG_MODE', 'false').lower() == 'true'
USE_OAUTH = os.getenv('USE_OAUTH', 'false').lower() == 'true'
GOOGLE_CLIENT_ID = os.getenv('GOOGLE_CLIENT_ID')
GOOGLE_CLIENT_SECRET = os.getenv('GOOGLE_CLIENT_SECRET')
SQLGLOT_DIALECT=os.getenv('SQLGLOT_DIALECT', 'snowflake')
